#ifndef Client_HPP
#define Client_HPP

#ifndef Q_MOC_RUN
#include <boost/noncopyable.hpp>
#include <boost/thread.hpp>
#include <boost/interprocess/sync/interprocess_semaphore.hpp>
#include <boost/regex.hpp>
#endif

#include <msgClient.hh>

#ifndef Q_MOC_RUN
#include <simple_logger.hpp>
#endif

#include <MsgStructure.hpp>

namespace hatptester{

class HATPClient{
private:
	//Name of the HATPOnboard client
	#define HATPName "HATP"

	const std::string name;
	bool exit;
	boost::mutex mutexExit; //Is used when changing the value of exit
	
	msgClient client;
	boost::mutex mutexClient;
	//boost::condition_variable conditionClient; //Is used to start the thread (says the client is ready)
	boost::interprocess::interprocess_semaphore clientSemaphore; //To notify the client that it can start
	
	std::string message;
	bool status;
	boost::mutex mutexMessageReceiver;
	boost::mutex mutexMessageReady;
	boost::condition_variable conditionMessageReady;
	
	bool hatpIsConnected;
	boost::mutex mutexhatpIsConnected;
	
	boost::thread thread;

	HATPClient();
	HATPClient(const HATPClient&);
	const HATPClient& operator=(const HATPClient&);
	
public:
	HATPClient(const std::string& name, const std::string& address, unsigned int port):
		name(name), exit(false), clientSemaphore(0), hatpIsConnected(false),
		thread(boost::bind(&HATPClient::threadMessageReceiver, this)){
		
		//Connect the client
		{
			boost::mutex::scoped_lock lock(mutexClient);
			client.connect(name, address, port);
		}

		clientSemaphore.post();
		
		if(!client.isConnected()){
			LOG(fatal)<<"Failed to connect to the server "<<address<<":"<<port;
			stopThread();
		}
	}
	
	bool isConnected(){
		boost::mutex::scoped_lock lock(mutexClient);
		return (!exit && client.isConnected());
	}
	
	bool sendMessage(const std::string& message){
		boost::mutex::scoped_lock lock(mutexClient);
		if(client.isConnected()){
			LOG(debug)<<"Send message : "<<message;
			if(client.sendMessage(HATPName, message)){
				return true;
			}else{
				boost::mutex::scoped_lock lock(mutexhatpIsConnected);
				hatpIsConnected=false;
				return false;
			}
		}
		return false;
	}
	
	std::pair<bool, std::string> requestValue(const std::string& request){//Sends a message to HATP and wait for its answer
		if(sendMessage(request)){
			bool success;
			//Get the message from he thread
			{
				boost::mutex::scoped_lock lock(mutexMessageReady);
				success = conditionMessageReady.timed_wait(lock, boost::posix_time::milliseconds(1000));
			}
			
			{
				if(success){
					boost::mutex::scoped_lock lock(mutexMessageReceiver);
					if(status){
						boost::mutex::scoped_lock lock(mutexhatpIsConnected);
						hatpIsConnected=true;
						return std::make_pair(true, message);
					}
				}else{
					LOG(warning)<<"Waited for a second but the message was still not ready, leave with error status";
					boost::mutex::scoped_lock lock(mutexhatpIsConnected);
					hatpIsConnected=false;
				}
			}
		}
		return std::make_pair(false, "");
		
	}
	
	bool ishatpConnected(){
		boost::mutex::scoped_lock lock(mutexhatpIsConnected);
		return (this->isConnected() && hatpIsConnected);
	}

private:
	bool isExitCommand(std::pair<std::string, std::string> message){
		if(message.first!=HATPName) return false;
		boost::smatch dummy;
		return boost::regex_match(message.second, dummy, exitCommandRegex);
	}
	
	bool isConnectCommand(std::pair<std::string, std::string> message){
		if(message.first!=HATPName) return false;
		boost::smatch dummy;
		return boost::regex_match(message.second, dummy, connectCommandRegex);
	}
	
	void threadMessageReceiver(){
		THREAD_NAME("Client");
	
		//Wait for connection
		clientSemaphore.wait();
	
		std::pair<std::string, std::string> messageTemp;
		bool localExit=false;
		
		{
			boost::mutex::scoped_lock lock(mutexClient);
			if(!client.isConnected()){
				LOG(fatal)<<"The client was not connected, the thread is stopped";
				localExit=true;
			}
		}
		
		while(!localExit) {
			bool sleep=false;
			mutexExit.lock();
			localExit=exit;
			mutexExit.unlock();
			
			if(!localExit){
				boost::mutex::scoped_lock lock(mutexClient);
				if(client.isConnected()){
					if(client.isMessageWaiting()){
						messageTemp = client.getMessage();
						if(messageTemp.first!=""){
							boost::mutex::scoped_lock lock(mutexMessageReceiver);
							message=messageTemp.second;
							status=true;
							LOG(debug)<<"Message received : "<<messageTemp.first<<" : \""<<message<<"\"";
							
							if(isExitCommand(messageTemp)){
								boost::mutex::scoped_lock lock(mutexhatpIsConnected);
								hatpIsConnected=false;
							}else if(isConnectCommand(messageTemp)){
								boost::mutex::scoped_lock lock(mutexhatpIsConnected);
								hatpIsConnected=true;
							}else{
								if(messageTemp.first==HATPName){
									boost::mutex::scoped_lock lock(mutexhatpIsConnected);
									hatpIsConnected=true;
								}
								conditionMessageReady.notify_one();
							}
						}else{
							boost::mutex::scoped_lock lock(mutexMessageReceiver);
							message="";
							status=false;
							if(messageTemp.second=="NOT_CONNECTED"){
								LOG(trace)<<"Message received indicates that the MsgServer stopped";
							
								localExit=true;
								
								boost::mutex::scoped_lock lock(mutexhatpIsConnected);
								hatpIsConnected=false;
							}
							conditionMessageReady.notify_one();
						}
						//Add all the cases : errors, quit, ...
					}else{
						sleep=true;
					}
				}else{
					LOG(trace)<<"Client is not connected, stop the client thread";
					localExit=true;
				}
			}
			
			if(sleep){
				usleep(200);
			}
		}
		
		LOG(trace)<<"The client's thread exits";
	}

public:
	void stopThread(){
		mutexExit.lock();
		bool localExit=exit;
		mutexExit.unlock();
		if(localExit==false){
			mutexExit.lock();
			exit=true;
			mutexExit.unlock();
			thread.join();
			LOG(trace)<<"Thread has joined, leave";
		}
	}
	
	~HATPClient(){
		stopThread();
	}
};

} //namespace

#endif //Client_HPP
