#include <View.hpp>

namespace hatptester{
//Default values
std::string HATPController::address("localhost");
unsigned int HATPController::port(5500);

MainWindow::MainWindow():QMainWindow(), 
	controller(HATPController::getInstance()),
	centralPart(this),
	taskBox("Task selection", this),
	taskView(&taskBox),
	parameterBox(this),
	defaultText("Parameters for the task : "),
	parametersView(this),
	informationsText("Please select a task (a method or an Action)\n\n\n\n\n\n", this),
	statusBarContent("", this),
	quitAction("&Quit", this), sendFirstAction("Request first plan", this), sendAllAction("&Request all plans", this),
	aboutAction("&About", this),
	maxTimeToggle("Set time limit to : ", this),
	maxTimeValue(this),
	changeToggle(false),
	maxTimeButton("Set", this),
	sendFirstPlanRequestButton("Request first plan", this),
	sendAllPlansRequestButton("Request all plans", this),
	timer(this)
	
{
	//First check if the controller is ready (its client is connected), if not leave
	if(!controller.isConnected()){ //Error message is generated by the client itself
		exit(99);
	}
	connect(&controller, SIGNAL(sendStop()), qApp, SLOT(quit()));
	connect(&controller, SIGNAL(sendStateChanged()), this, SLOT(stateChanged()));
	
	QTextCodec::setCodecForCStrings(QTextCodec::codecForName("UTF-8"));
	
	//Actions
	quitAction.setShortcut(QKeySequence("Ctrl+Q"));
	quitAction.setToolTip("Exit HATP Tester");
	quitAction.setIcon(QIcon::fromTheme("application-exit"));
	connect(&quitAction, SIGNAL(triggered()), qApp, SLOT(quit()));
	
	sendAllAction.setShortcut(QKeySequence("Ctrl+S"));
	sendAllAction.setToolTip("Request all plans");
	sendAllAction.setEnabled(false);
//TODO	sendAllAction.setIcon(QIcon::fromTheme(""));
	connect(&sendAllAction, SIGNAL(triggered()), this, SLOT(sendAllPlansRequest()));
	
	sendFirstAction.setShortcut(QKeySequence("Ctrl+Shift+S"));
	sendFirstAction.setToolTip("Request first plan");
	sendFirstAction.setEnabled(false);
//TODO	sendFirstAction.setIcon(QIcon::fromTheme(""));
	connect(&sendFirstAction, SIGNAL(triggered()), this, SLOT(sendFirstPlanRequest()));
	
	connect(&aboutAction, SIGNAL(triggered()), this, SLOT(about()));
	
	//Menu
	QMenu *fileMenu = menuBar()->addMenu("&File");
	fileMenu->addAction(&sendAllAction);
	fileMenu->addAction(&sendFirstAction);
	fileMenu->addSeparator();
	fileMenu->addAction(&quitAction);
	
	QMenu *helpMenu = menuBar()->addMenu("&Help");
	helpMenu->addAction(&aboutAction);
	
	//Top left part (Tasks' list)
	taskView.setModel(&controller.getTasks());
	taskView.setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
	taskView.setEditTriggers(QAbstractItemView::NoEditTriggers);
	
	QVBoxLayout* topLeftBox=new QVBoxLayout(&taskBox);
	topLeftBox->addWidget(&taskView);
	taskBox.setLayout(topLeftBox);
	
	//Top right part (Parameters for the selected task)
	QVBoxLayout* topRightBox=new QVBoxLayout(&parameterBox);
	QStandardItemModel* model=new QStandardItemModel(5, 2, this);
	QStringList labels;
	labels<<"Name"<<"Value";
	model->setHorizontalHeaderLabels(labels);
	QStringList typeLabels;
	typeLabels<<"Type"<<"Type"<<"Type"<<"Type"<<"Type";
	model->setVerticalHeaderLabels(typeLabels);
	parametersView.setModel(model);
	ParameterDelegate* delegate=new ParameterDelegate(this);
	connect(delegate, SIGNAL(requestIsReady()), this, SLOT(activateSend()));
	connect(delegate, SIGNAL(modelUpdated()), this, SLOT(updateInformations()));
	parametersView.setItemDelegate(delegate);
	parametersView.setSelectionMode(QAbstractItemView::SingleSelection);
	parametersView.setEditTriggers(QAbstractItemView::AllEditTriggers);
	parametersView.horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
	parametersView.horizontalHeader()->setStretchLastSection(true);
	parametersView.setDisabled(true);
	oldParametersView=false;
	topRightBox->addWidget(&parametersView);
	
	//TODO Use the metric for both to prevent some weird cases
	std::string longestName=defaultText+controller.getLongestTaskName();
	QFontMetrics metric(QFont("Times", 14, QFont::Bold));
	QRect rect(metric.boundingRect(longestName.c_str()));
	parameterBox.setMinimumWidth(rect.width());
	
	parameterBox.setTitle(defaultText.c_str());
	parameterBox.setLayout(topRightBox);
	
	//Bottom left part (Informations)
	QGroupBox* informationsBox=new QGroupBox("Informations", this);
	QVBoxLayout* bottomLeftLayout=new QVBoxLayout(informationsBox);
	informationsText.setWordWrap(true);
	informationsText.setMinimumSize(informationsText.sizeHint());
	bottomLeftLayout->addWidget(&informationsText);
	informationsBox->setLayout(bottomLeftLayout);
	
	//Bottom right part (Send button and Time Limit)
	sendFirstPlanRequestButton.setToolTip("Send the request and stop after first plan is found");
	sendFirstPlanRequestButton.setEnabled(false);
	sendFirstPlanRequestButton.setMinimumSize(sendFirstPlanRequestButton.sizeHint());
	connect(&sendFirstPlanRequestButton, SIGNAL(clicked()), this, SLOT(sendFirstPlanRequest()));
	sendAllPlansRequestButton.setToolTip("Send the request and computes all possible plans");
	sendAllPlansRequestButton.setEnabled(false);
	sendAllPlansRequestButton.setMinimumSize(sendAllPlansRequestButton.sizeHint());
	connect(&sendAllPlansRequestButton, SIGNAL(clicked()), this, SLOT(sendAllPlansRequest()));

	connect(&maxTimeToggle, SIGNAL(stateChanged(int)), this, SLOT(maxTimeToggled(int)));
	maxTimeValue.setMinimum(0);
	maxTimeValue.setMaximum(3600);
	maxTimeValue.setSingleStep(10);
	maxTimeValue.setSuffix(" - unlimited");
	maxTimeValue.setValue(0);
	maxTimeValue.setEnabled(false);
	maxTimeValue.setMinimumSize(maxTimeValue.sizeHint());
	maxTimeValue.setMaximumSize(maxTimeValue.sizeHint());
	connect(&maxTimeValue, SIGNAL(valueChanged(int)), this, SLOT(maxTimeChanged(int)));
	maxTimeButton.setToolTip("Send time limit request to HATP");
	maxTimeButton.setEnabled(false);
	oldMaxTimeButton=false;
	connect(&maxTimeButton, SIGNAL(clicked()), this, SLOT(sendMaxTime()));
	
	QGridLayout* maxTimeLayout=new QGridLayout;
	//No need for parent, thanks to the following centralPartLayout->addLayout
	maxTimeLayout->addWidget(&maxTimeToggle, 0, 1);
	maxTimeLayout->addWidget(&maxTimeValue, 0, 2);
	maxTimeLayout->addWidget(&maxTimeButton, 0, 3);
	maxTimeLayout->addWidget(&sendFirstPlanRequestButton, 2, 2);
	maxTimeLayout->addWidget(&sendAllPlansRequestButton, 2, 3);

	QGridLayout* centralPartLayout=new QGridLayout;
	//No need for parent because it becomes the central widget and the window takes ownership
	centralPartLayout->addWidget(&taskBox, 0, 0);
	centralPartLayout->addWidget(&parameterBox, 0, 1);
	centralPartLayout->addWidget(informationsBox, 1, 0);
	centralPartLayout->addLayout(maxTimeLayout, 1, 1);

	centralPart.setLayout(centralPartLayout);

	this->setMinimumSize(centralPart.sizeHint());
	setCentralWidget(&centralPart);
	
	statusBar()->addWidget(&statusBarContent);
	
	//Finish by starting the timer
	connect(&timer, SIGNAL(timeout()), this, SLOT(periodicCheck()));
	timer.start(200);
	
	if(controller.isReady()){
		controllerWasReady=true;
		restoreAllButtonState();
	}else{
		controllerWasReady=false;
		setAllButtonsDisable();
	}
	
	show();
}

void MainWindow::setAllButtonsDisable(){
	sendFirstAction.setDisabled(true);
	sendAllAction.setDisabled(true);
	sendFirstPlanRequestButton.setDisabled(true);
	sendAllPlansRequestButton.setDisabled(true);
	maxTimeToggle.setDisabled(true);
	maxTimeValue.setDisabled(true);
	maxTimeButton.setDisabled(true);
	taskView.setDisabled(true);
	disconnect(this, SLOT(clicTask(const QModelIndex&)));
	parametersView.setDisabled(true);
	statusBarContent.setText("HATP is NOT connected");
}

void MainWindow::restoreAllButtonState(){
	sendAllAction.setEnabled(controller.isReadyCurrentTask());
	sendFirstAction.setEnabled(controller.isReadyCurrentTask());
	sendFirstPlanRequestButton.setEnabled(controller.isReadyCurrentTask());
	sendAllPlansRequestButton.setEnabled(controller.isReadyCurrentTask());
	maxTimeToggle.setEnabled(true);
	maxTimeValue.setEnabled(oldMaxTimeButton);
	maxTimeButton.setEnabled(oldMaxTimeButton);
	taskView.setEnabled(true);
	connect(&taskView, SIGNAL(clicked(const QModelIndex&)), this, SLOT(clicTask(const QModelIndex&)));
	parametersView.setEnabled(oldParametersView);
	statusBarContent.setText("HATP is connected");
}

void MainWindow::resizeEvent(QResizeEvent* event){
	centralPart.resize(event->size());
}

//SLOTS
void MainWindow::about(){
	QMessageBox::about(this, "About hatptester", "<H1> hatptester </H1>"
		"The <b>hatptester</b> aims to provide a simple interface to "
		"send requests to HATP Symbolic Planner.<br /><br />"
		"created by Raphaël Lallement < br />"
		"<i>raphael.lallement@laas.fr</i>"
	);
}

void MainWindow::sendFirstPlanRequest(){
	controller.sendFirstPlanRequest();
}

void MainWindow::sendAllPlansRequest(){
	controller.sendAllPlansRequest();
}

void MainWindow::activateSend(){
	sendAllAction.setEnabled(true);
	sendFirstAction.setEnabled(true);
	sendFirstPlanRequestButton.setEnabled(true);
	sendAllPlansRequestButton.setEnabled(true);
}

void MainWindow::updateInformations(){
	int jumps=5; //Represents the number of line to jump at the end to get a correct shape
	std::string informations;
	if(controller.isReadyCurrentTask()){
		informations="Request can be sent\n";
		
	}else{
		std::set<std::string> missingParameters=controller.getMissingParameters(controller.getCurrentTaskName());
		informations="To send the request, please set a value to the following parameters :\n";
		--jumps;
		for(std::set<std::string>::const_iterator it=missingParameters.begin(); it!=missingParameters.end(); ++it){
			informations+="\t"+*it+"\n";
			--jumps;
		}
	}
	while(jumps>0){
		informations+="\n";
		--jumps;
	}
	informationsText.setText(informations.c_str());
}

void MainWindow::clicTask(const QModelIndex& index)
{
	QVariant selectedTask(controller.getTasks().data(index, Qt::DisplayRole));
	std::string task(selectedTask.toString().toStdString());
	if(task=="" || task==" -- Methods : " || task==" -- Actions : "){
		return;
	}
	std::string title=defaultText+task;
	parameterBox.setTitle(title.c_str());
	
	QStandardItemModel* parameters=controller.getTaskParameter(task);
	if(parameters==NULL){
		parameters=new QStandardItemModel(5, 2, this);
		QStringList labels;
		labels<<"Name"<<"Value";
		parameters->setHorizontalHeaderLabels(labels);
		QStringList typeLabels;
		typeLabels<<"Type"<<"Type"<<"Type"<<"Type"<<"Type";
		parameters->setVerticalHeaderLabels(typeLabels);
		sendAllAction.setEnabled(false);
		sendFirstAction.setEnabled(false);
		sendFirstPlanRequestButton.setEnabled(false);
		sendAllPlansRequestButton.setEnabled(false);
		informationsText.setText("Invalid task !");
		parametersView.setDisabled(true);
		oldParametersView=false;
	}else{
		parametersView.setEnabled(true);
		oldParametersView=true;
		sendAllAction.setEnabled(controller.isReadyTask(task));
		sendFirstAction.setEnabled(controller.isReadyTask(task));
		sendFirstPlanRequestButton.setEnabled(controller.isReadyTask(task));
		sendAllPlansRequestButton.setEnabled(controller.isReadyTask(task));
		updateInformations();
	}
	parametersView.setModel(parameters);
	parametersView.horizontalHeader()->setStretchLastSection(true);
}

void MainWindow::stateChanged(){
	std::string title=defaultText+"";
	parameterBox.setTitle(title.c_str());
	
	QStandardItemModel* parameters=new QStandardItemModel(5, 2, this);
	QStringList labels;
	labels<<"Name"<<"Value";
	parameters->setHorizontalHeaderLabels(labels);
	QStringList typeLabels;
	typeLabels<<"Type"<<"Type"<<"Type"<<"Type"<<"Type";
	parameters->setVerticalHeaderLabels(typeLabels);
	sendAllAction.setEnabled(false);
	sendFirstAction.setEnabled(false);
	sendFirstPlanRequestButton.setEnabled(false);
	sendAllPlansRequestButton.setEnabled(false);
	if(controller.isReady()){
		informationsText.setText("Please select a task (a method or an Action)");
		parametersView.setEnabled(true);
	}else{
		informationsText.setText("Error while initialising the interface");
		parametersView.setEnabled(false);
	}
	oldParametersView=false;
	parametersView.setModel(parameters);
	parametersView.horizontalHeader()->setStretchLastSection(true);
}

void MainWindow::periodicCheck(){
	if(!controller.isConnected()){ //Either a problem occured or the client underneath disconnected form the server
		LOG(error)<<"Exiting hatptester (reason: MsgServer was killed)";
		qApp->quit();
	}
	if(controllerWasReady && !controller.ishatpConnected()){
		controllerWasReady=false;
		setAllButtonsDisable();
	}else if(!controllerWasReady && controller.ishatpConnected()){
		controllerWasReady=true;
		controller.reloadModel();
		restoreAllButtonState();
	}
}

void MainWindow::maxTimeToggled(int toggle){
	changeToggle=!changeToggle;
	maxTimeButton.setEnabled(changeToggle);
	oldMaxTimeButton=changeToggle;
	maxTimeChanged(toggle);
}

void MainWindow::maxTimeChanged(int change){
	if(maxTimeToggle.isChecked()){
		maxTimeValue.setMinimum(1);
		maxTimeValue.setSuffix(" s");
		maxTimeValue.setEnabled(true);
		maxTimeButton.setEnabled(true);
		oldMaxTimeButton=true;
	}else{
		maxTimeValue.setMinimum(0);
		maxTimeValue.setValue(0);
		maxTimeValue.setSuffix(" - unlimited");
		maxTimeValue.setEnabled(false);
	}
}

void MainWindow::sendMaxTime(){
	changeToggle=false;
	maxTimeButton.setEnabled(false);
	oldMaxTimeButton=false;
	if(maxTimeToggle.isChecked()){
		controller.sendMaxTime(maxTimeValue.value());
	}else{
		controller.sendMaxTime(0);
	}
}

}; //End namespace
