#include <iostream>
#include <csignal>

#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>

#include <QApplication>

#include <View.hpp>
#include <Controller.hpp>

#define DEFAULT_ADDRESS "localhost"
#define DEFAULT_PORT 5500

void signalHandler(int s){
	LOG(fatal)<<"Caught signal \""<<strsignal(s)<<"\"";
	
	hatptester::HATPController::getInstance().exit();
}

int main(int argc, char ** argv){
	THREAD_NAME("Main");

	std::string address (DEFAULT_ADDRESS);
	unsigned int port (DEFAULT_PORT);
	
	//Supported options
	boost::program_options::options_description desc("Allowed options");

	desc.add_options()
		("help,h", "Produce help message")
		("address,a", boost::program_options::value<std::string>(&address), "The address of the server (default : localhost)")
		("port,p", boost::program_options::value<unsigned int>(&port), "The port number of the server (default : 5500)");
	
	//The order for parameter (without the -.. options) is : address port
	boost::program_options::positional_options_description positional;
	positional.add("address", 1);
	positional.add("port", 1);

	//Parse the parameters
	boost::program_options::variables_map variables;
	try{
		boost::program_options::store(boost::program_options::command_line_parser(argc, argv).options(desc).positional(positional).run(), variables);
	}catch(boost::program_options::validation_error e){
		//Called if a parameter get a value that does not match its type
		LOG(error)<<"Catch a validation error exception : "<<e.what();
		exit(99);
	}catch(boost::program_options::unknown_option e){
		//Called if an unknown option is given (e.g. : -z)
		LOG(warning)<<"Catch an excpetion : "<<e.what();
	}
	boost::program_options::notify(variables);
	
	//Print usage
	if(variables.count("help")){
		std::string software(argv[0]);
		std::cout<<"Usage:\tDefault:\t"<<software<<std::endl;
		std::cout<<"\tCompact:\t"<<software<<" address:port"<<std::endl;
		std::cout<<"\tStandard:\t"<<software<<" address port"<<std::endl;
		std::cout<<"\tComplete:\t"<<software<<" [-a address] [-p port]"<<std::endl<<std::endl;
		std::cout<<desc<<std::endl;
		return 0;
	}
	
	//Allow to write "address:port" (in addition to "address port")
	if(address.find(":")!=std::string::npos){
		//The address contains the port number as well
		size_t pos=address.find(":");
		
		std::string tempPort=address.substr(pos+1);
		if(!tempPort.empty()){
			try{
				port=boost::lexical_cast<unsigned int>(tempPort);
			}catch(const boost::bad_lexical_cast &){
				LOG(error)<<"The port must be a number !";
				exit(99);
			}
		}
		
		if(pos!=0){
			address=address.substr(0, pos);
		}else{
			address="localhost";
		}
	}
	
	LOG(info)<<argv[0]<<" starts";
	LOG(info)<<"It will try to connect to: "<<address<<":"<<port;
	
	hatptester::HATPController::setInstance(address, port);
	
	int fakeArgc=1; //To change if one want to use the possibilities of Qt
	char *fakeArgv[1]={argv[0]};
	QApplication app(fakeArgc, fakeArgv);

	hatptester::MainWindow window;
	
	//Set the signal handler to catch ctrl-c (i.e. SIGINT) and other
	struct sigaction sigHandler;
	sigHandler.sa_handler = signalHandler;
	sigemptyset(&sigHandler.sa_mask);
	sigHandler.sa_flags = 0;
	sigaction(SIGINT, &sigHandler, NULL);
	
	return app.exec();
}
